ROLLBAR_SERVER_POST_KEY = '62f87c5df3674629b143a137de3d3244'

SEGMENT_IO_WRITE_KEY_PROD = '51ybGTeFEFU1xo6u10XMDrr6kATFyRyh'
SEGMENT_IO_CLI_EVENT = 'dcos-cli'
SEGMENT_IO_CLI_ERROR_EVENT = 'dcos-cli-error'
SEGMENT_URL = 'https://api.segment.io/v1'

DCOS_PRODUCTION_ENV = 'DCOS_PRODUCTION'
